#!/usr/bin/python3

import nsysstats

class PIXSummary(nsysstats.Report):

    usage = f"""{{SCRIPT}} -- PIX Range Summary

    No arguments.

    Output: All time values default to nanoseconds
        Time : Percentage of "Total Time"
        Total Time : Total time used by all instances of this range
        Instances: Number of instances of this range
        Avg : Average execution time of this range
        Med : Median execution time of this rage
        Min : Smallest execution time of this range
        Max : Largest execution time of this range
        StdDev : Standard deviation of execution time of this range
        Range : Name of the range

    This report provides a summary of PIX CPU debug markers,
    and their execution times. Note that the "Time" column
    is calculated using a summation of the "Total Time" column, and represents
    that range's percent of the execution time of the ranges listed, and not a
    percentage of the application wall or CPU execution time.
"""

    query = f"""
WITH
    beginEvents AS (
        SELECT
            *
        FROM
            PIX_DEBUG_API
        WHERE textId IS NOT NULL
    ),
    endEvents AS (
        SELECT
            *
        FROM
            PIX_DEBUG_API
        WHERE textId IS NULL
    ),
    summary AS (
        SELECT
            beginEvents.textId AS textId,
            sum(endEvents.start - beginEvents.end) AS total,
            count(*) AS num,
            avg(endEvents.start - beginEvents.end) AS avg,
            median(endEvents.start - beginEvents.end) AS med,
            min(endEvents.start - beginEvents.end) AS min,
            max(endEvents.start - beginEvents.end) AS max,
            stdev(endEvents.start - beginEvents.end) AS stddev
        FROM
            beginEvents
        INNER JOIN
            endEvents
            ON beginEvents.correlationId == endEvents.correlationId
        GROUP BY 1
    ),
    totals AS (
        SELECT sum(total) AS total
        FROM summary
    )

SELECT
    round(summary.total * 100.0 / (SELECT total FROM totals), 1) AS "Time:ratio_%",
    summary.total AS "Total Time:dur_ns",
    summary.num AS "Instances",
    round(summary.avg, 1) AS "Avg:dur_ns",
    round(summary.med, 1) AS "Med:dur_ns",
    summary.min AS "Min:dur_ns",
    summary.max AS "Max:dur_ns",
    round(summary.stddev, 1) AS "StdDev:dur_ns",
    ids.value AS "Range"
FROM
    summary
LEFT JOIN
    StringIds AS ids
    ON ids.id == summary.textId
WHERE summary.total > 0
ORDER BY 2 DESC
;
"""

    table_checks = {
        'PIX_DEBUG_API':
            "{DBFILE} does not contain DX11/DX12 CPU debug markers."
    }

if __name__ == "__main__":
    PIXSummary.Main()
